use piet::kurbo::Rect;
use serde_with::DeserializeFromStr;
use std::fmt::{Display, Formatter, Write};
use std::str::FromStr;
use thiserror::Error;

#[derive(Debug, Clone)]
pub enum ImageTransform {
    Invert,
    Brighten(f32),
    Crop(Rect),
    HorizontalFlip,
    VerticalFlip,
    Rotate { quarter_clockwise_turns: u8 },
}

impl Display for ImageTransform {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        match self {
            ImageTransform::Invert => f.write_str("invert"),
            ImageTransform::Brighten(value) => f.write_fmt(format_args!("brighten={}", value)),
            ImageTransform::Crop(_) => f.write_str("crop=<...>"),
            ImageTransform::HorizontalFlip => f.write_str("horizontal_flip"),
            ImageTransform::VerticalFlip => f.write_str("vertical_flip"),
            ImageTransform::Rotate { quarter_clockwise_turns } => f.write_fmt(format_args!("rotate={}", quarter_clockwise_turns)),
        }?;

        Ok(())
    }
}

#[derive(DeserializeFromStr)]
pub struct ImageFilter {
    scale: f32,
    alpha: f32,
    blur: f32,
    transforms: Vec<ImageTransform>,
}

#[derive(Debug, Error)]
pub enum ImageFilterFromStringError {
    #[error("Unknown filter: {name}")]
    UnknownFilter { name: String },

    #[error("Filter {name} can only be used once.")]
    SingletonSpecifiedMoreThanOnce { name: String },

    #[error("All singleton filters must be placed before any transforms, but {singleton} was placed after {transform}.")]
    SingletonSpecifiedAfterTransform { singleton: String, transform: ImageTransform },

    #[error("{quarter_clockwise_turns} 90° turns can be simplified.")]
    PointlessRotation { quarter_clockwise_turns: u8 },

    #[error("Filter {filter_name} requires a value ({filter_name}=<value>).")]
    FilterValueMissing { filter_name: String },

    #[error("The following value supplied to {filter_name} could not be parsed: {raw_value}")]
    FilterValueNotParsable { filter_name: String, raw_value: String },

    #[error("{value} is not in the range ({range}) required by {filter_name}.")]
    FilterValueNotInRange { filter_name: String, value: String, range: String },
}

impl FromStr for ImageFilter {
    type Err = ImageFilterFromStringError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let filters: Vec<&str> = s.split('|').map(|f| f.trim()).collect();

        let scale: Option<f32> = None;
        let alpha: Option<f32> = None;
        let blur: Option<f32> = None;

        let mut transforms: Vec<ImageTransform> = Vec::new();

        for filter in filters {
            let split_filter = filter.split_once('=');
            let (filter_name, optional_raw_value) = if let Some((filter_name, raw_value)) = split_filter {
                (filter_name.to_owned(), Some(raw_value.to_owned()))
            } else {
                (filter.to_owned(), None)
            };

            let use_raw_value = || -> Result<String, ImageFilterFromStringError> {
                optional_raw_value.ok_or_else(|| ImageFilterFromStringError::FilterValueMissing {
                    filter_name: filter_name.clone(),
                })
            };

            match filter_name.as_str() {
                "scale" | "alpha" | "blur" => {
                    if let Some(transform) = transforms.first() {
                        return Err(ImageFilterFromStringError::SingletonSpecifiedAfterTransform {
                            singleton: filter_name.clone(),
                            transform: transform.clone(),
                        });
                    }

                    #[allow(clippy::type_complexity)]
                    let (mut variable, check_range, range_string): (Option<f32>, Box<dyn Fn(&f32) -> bool>, &'static str) = match filter_name.as_str() {
                        "scale" => (scale, Box::new(|v| (0.0..=10.0_f32).contains(v)), "0..=10"),
                        "alpha" => (alpha, Box::new(|v| (0.0..1.0_f32).contains(v)), "0..<1"),
                        "blur" => (blur, Box::new(|v| (0.0..=10.0_f32).contains(v)), "0..=10"),
                        _ => unreachable!("filter_name is narrowed down in the outer match statement"),
                    };

                    if variable.is_some() {
                        return Err(ImageFilterFromStringError::SingletonSpecifiedMoreThanOnce { name: filter_name.clone() });
                    }

                    let raw_value = use_raw_value()?;
                    let value = f32::from_str(&raw_value).map_err(|_| ImageFilterFromStringError::FilterValueNotParsable {
                        filter_name: filter_name.clone(),
                        raw_value,
                    })?;

                    if !check_range(&value) {
                        return Err(ImageFilterFromStringError::FilterValueNotInRange {
                            filter_name: filter_name.clone(),
                            range: range_string.to_owned(),
                            value: value.to_string(),
                        });
                    }

                    let _ = variable.insert(value);
                }
                _ => return Err(ImageFilterFromStringError::UnknownFilter { name: filter_name }),
            }
        }

        Ok(ImageFilter {
            scale: scale.unwrap_or(1.0),
            alpha: alpha.unwrap_or(1.0),
            blur: blur.unwrap_or(0.0),
            transforms,
        })
    }
}
